//
//  View.m
//  Queue
//
//  Created by NYU User on 9/8/10.
//  Copyright 2010 __MyCompanyName__. All rights reserved.
//

#import <AudioToolbox/AudioToolbox.h>
//#import <Foundation/Foundation.h>
#import "View.h"

#define KILOBYTE 1024	//number of bytes in a kilobyte
#define numberInputBuffers  ((size_t)3)
#define numberOutputBuffers ((size_t)3)

//Structure passed to inputCallback.

struct inputUserData {
	AudioStreamBasicDescription format;
	AudioQueueRef queue;
	AudioQueueBufferRef buffer[numberInputBuffers];
	AudioFileID fileID;
	UInt32 bufferByteSize;
	SInt64 currentPacket;
	bool isRunning;
};

//Structure passed to outputCallback.

struct outputUserData {
	AudioStreamBasicDescription format;
	AudioQueueRef queue;
	AudioQueueBufferRef buffer[numberOutputBuffers];
	AudioFileID fileID;
	UInt32 bufferByteSize;
	SInt64 currentPacket;
	bool isRunning;
	
	UInt32 numPacketsToRead;
	AudioStreamPacketDescription *description;
};

static struct inputUserData inputData;
static struct outputUserData outputData;

static NSString *fileName = @"soundfile.aif";
static AudioFileTypeID fileType = kAudioFileAIFFType;

//An input audio queue calls the following function repeatedly
//to copy a buffer into a file.

static void inputCallback(
	void *userData,
	AudioQueueRef queue,
	AudioQueueBufferRef buffer,
	const AudioTimeStamp *startTime,
	UInt32 numPackets,
	const AudioStreamPacketDescription *description
) {
	struct inputUserData *p = (struct inputUserData *)userData;
	
	if (numPackets == 0 && p->format.mBytesPerPacket > 0) {
		numPackets =
			buffer->mAudioDataByteSize / p->format.mBytesPerPacket;
	}
	
	OSStatus status = AudioFileWritePackets(
		p->fileID,
		false,
		buffer->mAudioDataByteSize,
		description,
		p->currentPacket,
		&numPackets,
		buffer->mAudioData
	);
	
	if (status != noErr) {
		NSLog(@"AudioFileWritePackets returned %d", status);
		return;
	}
	
	p->currentPacket += numPackets;
	if (p->isRunning) {
		status = AudioQueueEnqueueBuffer(p->queue, buffer, 0, NULL);
		if (status != noErr) {
			NSLog(@"inputCallback: AudioQueueEnqueueBuffer returned %d", status);
		}
	}
}

//An output audio queue calls the following function repeatedly
//to copy data from a file into a buffer.

static void outputCallback(
	void *userData,
	AudioQueueRef queue,
	AudioQueueBufferRef buffer
) {
	struct outputUserData *p = (struct outputUserData *)userData;
	if (!p->isRunning) {
		return;
	}
	
	UInt32 numBytesReadFromFile;
	UInt32 numPackets = p->numPacketsToRead;
	
	OSStatus status = AudioFileReadPackets(
		p->fileID,
		false,
		&numBytesReadFromFile,
		p->description,
		p->currentPacket,
		&numPackets,
		buffer->mAudioData
	);
	
	if (status != noErr) {
		NSLog(@"AudioFileReadPackets returned %d", status);
	}
	
	if (numPackets == 0) {
		status = AudioQueueStop(p->queue, false);
		if (status != noErr) {
			NSLog(@"AudioQueueStop returned %d", status);
		}
		p->isRunning = false;
	} else {
		buffer->mAudioDataByteSize = numBytesReadFromFile;
		status = AudioQueueEnqueueBuffer(
			p->queue,
			buffer,
			p->description == NULL ? 0 : numPackets,
			p->description
		);
		if (status != noErr) {
			NSLog(@"outputCallback: AudioQueueEnqueueBuffer returned %d", status);
		}
		p->currentPacket += numPackets;
	}
}

static void inputBufferSize(
	AudioQueueRef audioQueue,
	AudioStreamBasicDescription *description,
	Float64 seconds,
	UInt32 *bufferSize
) {
	UInt32 maxPacketSize = description->mBytesPerPacket;
	if (maxPacketSize == 0) {
		UInt32 maxVBRPacketSize = sizeof maxPacketSize;
		OSStatus status = AudioQueueGetProperty(
			audioQueue,
			kAudioQueueProperty_MaximumOutputPacketSize,
			&maxPacketSize,
			&maxVBRPacketSize
		);
		if (status != noErr) {
			NSLog(@"AudioQueueGetProperty returned %d", status);
		}
	}
	
	*bufferSize = description->mSampleRate * maxPacketSize * seconds;
	*bufferSize = MIN(*bufferSize, 320 * KILOBYTE);
}

static void outputBufferSize(
	AudioStreamBasicDescription *ASBDesc,
	UInt32 maxPacketSize,
	Float64 seconds,
	UInt32 *bufferSize,
	UInt32 *numPacketsToRead
) {
	static const UInt32 maxBufferSize = 320 * KILOBYTE;
	
	if (ASBDesc->mFramesPerPacket == 0) {
		*bufferSize = MAX(maxBufferSize, maxPacketSize);
	} else {
		*bufferSize = seconds * maxPacketSize * ASBDesc->mSampleRate
			/ ASBDesc->mFramesPerPacket;
	}
	
	if (*bufferSize > maxBufferSize
		&& *bufferSize > maxPacketSize) {
		*bufferSize = maxBufferSize;
	} else {
		*bufferSize = MAX(*bufferSize, 16 * KILOBYTE);
	}
	
	*numPacketsToRead = *bufferSize / maxPacketSize;
}

//Some audio formats require magic cookies, although our format (AIFF)
//does not.
//Copy the magic cookie from an input audio queue to an audio file
//into which we are recording.

OSStatus SetMagicCookieForInputFile(AudioQueueRef inQueue,
	AudioFileID inFile) {
	
	UInt32 cookieSize;
	OSStatus result = AudioQueueGetPropertySize(inQueue,
		kAudioQueueProperty_MagicCookie, &cookieSize);
	if (result != noErr) {
		NSLog(@"AudioQueueGetPropertySize("
			  "kAudioQueueProperty_MagicCookie) returned %d", result);
		return result;
	}
	
	char *const magicCookie = malloc(cookieSize);
	if (magicCookie == NULL) {
		NSLog(@"malloc failed");
		return noErr + 1;
	}
	
	result = AudioQueueGetProperty(inQueue,
		kAudioQueueProperty_MagicCookie, magicCookie, &cookieSize);
	
	if (result != noErr) {
		NSLog(@"AudioQueueGetProperty(kAudioQueueProperty_MagicCookie) "
			  "returned %d", result);
		return result;
	}
	
	result = AudioFileSetProperty(inFile,
		kAudioFilePropertyMagicCookieData, cookieSize, magicCookie);
	free(magicCookie);
	return result;
}

//Copy the magic cookie from an audio file that is being played
//to an output audio queue.

OSStatus SetMagicCookieForOutputQueue(AudioQueueRef inQueue,
	AudioFileID inFile) {
	
	UInt32 cookieSize = sizeof (UInt32);
	OSStatus status = AudioFileGetPropertyInfo(
		inFile,
		kAudioFilePropertyMagicCookieData,
		&cookieSize,
		NULL
	);
	
	if (status != noErr || cookieSize == 0) {
		NSLog(@"AudioFileGetPropertyInfo returned %d, cookieSize == %u",
			  status, cookieSize);
	}
	
	char *magicCookie = (char *)malloc(cookieSize);
	if (magicCookie == NULL) {
		NSLog(@"malloc failed");
	}
	
	status = AudioFileGetProperty(
		inFile,
		kAudioFilePropertyMagicCookieData,
		&cookieSize,
		magicCookie
	);
	
	if (status != noErr) {
		NSLog(@"AudioFileGetProperty returned %d", status);
	}
	
	status = AudioQueueSetProperty(
		inQueue,
		kAudioQueueProperty_MagicCookie,
		magicCookie,
		cookieSize
	);
	
	if (status != noErr) {
		NSLog(@"AudioQueueSetProperty returned %d", status);
	}
	
	free(magicCookie);
	return status;
}

@implementation View

- (id) initWithFrame: (CGRect) frame {
	if ((self = [super initWithFrame: frame]) != nil) {
		// Initialization code
		self.backgroundColor = [UIColor yellowColor];
		
		//Create the CFURLRef.
		
		const char *const pathname = [[NSHomeDirectory()
			stringByAppendingPathComponent: fileName]
			cStringUsingEncoding: NSUTF8StringEncoding];
		
		//NSLog(@"pathname == \"%s\"", pathname);
		
		url = CFURLCreateFromFileSystemRepresentation(
			kCFAllocatorDefault,
			(const UInt8 *)pathname,
			strlen(pathname),
			false															
		);
		
		NSLog(@"audioFileURL == \"%@\"", url);
		
		//Create the segmented control.
		
		NSArray *items = [NSArray arrayWithObjects:
			@"Record",
			@"Stop",
			@"Play",
			nil
		];
		
		control = [[UISegmentedControl alloc] initWithItems: items];
		
		//Center the control in the SegmentedView.
		CGSize s = self.bounds.size;
		self.bounds = CGRectMake(
			-s.width / 2,
			-s.height / 2,
			s.width,
			s.height
		);
		control.center = CGPointZero;
		
		//Can't play until we have recorded something.
		[control setEnabled: NO forSegmentAtIndex: 2];
		
		[control addTarget: self
			action: @selector(valueChanged:)
			forControlEvents: UIControlEventValueChanged];
		
		//Center the recording light above the recording button.
		CGFloat length = 10;
		s = control.bounds.size;
		CGRect frame = CGRectMake(
			-s.width / control.numberOfSegments - length / 2,
			-s.height / 2 - 2 * length,
			length,
			length
		);
		light = [[UIView alloc] initWithFrame: frame];
		light.backgroundColor = [UIColor grayColor];
		
		[self addSubview: control];
		[self addSubview: light];
	}
	return self;
}

- (void) stop {
	if (inputData.isRunning) {
		inputData.isRunning = false;
		
		OSStatus status = AudioQueueStop(inputData.queue, true);
		if (status != noErr) {
			NSLog(@"AudioQueueStop returned %d", status);
		}
		light.backgroundColor = [UIColor grayColor];
		
		status = AudioQueueDispose(inputData.queue, true);
		if (status != noErr) {
			NSLog(@"AudioQueueDispose returned %d", status);
		}
		
		status = AudioFileClose(inputData.fileID);
		if (status != noErr) {
			NSLog(@"AudioFileClose returned %d", status);
		}
	}
	
	if (outputData.isRunning) {
		outputData.isRunning = false;
		
		OSStatus status = AudioQueueDispose(outputData.queue, true);
		if (status != noErr) {
			NSLog(@"AudioQueueDispose returned %d", status);
		}
		
		status = AudioFileClose(outputData.fileID);
		if (status != noErr) {
			NSLog(@"AudioFileClose returned %d", status);
		}
		
		free(outputData.description);
	}
}

- (void) record {
	[self stop];

	//Set up an audio format for recording, p. 28.
	
	inputData.format.mFormatID = kAudioFormatLinearPCM;
	inputData.format.mSampleRate = 44100;		//210 * 210
	inputData.format.mChannelsPerFrame = 2;	//stereo
	inputData.format.mBitsPerChannel = 16;
	inputData.format.mFramesPerPacket = 1;
	
	//CHAR_BIT is the number of bits in a byte.
	inputData.format.mBytesPerFrame =
		inputData.format.mBitsPerChannel
		* inputData.format.mChannelsPerFrame / CHAR_BIT;

	inputData.format.mBytesPerPacket =
		inputData.format.mBytesPerFrame
		* inputData.format.mFramesPerPacket;
	
	inputData.format.mFormatFlags =
		kLinearPCMFormatFlagIsBigEndian
		| kLinearPCMFormatFlagIsSignedInteger
		| kLinearPCMFormatFlagIsPacked;
	
	//Create a recording audio queue, p. 29.
	
	OSStatus status = AudioQueueNewInput(
		&inputData.format,
		inputCallback,
		&inputData,
		NULL,
		kCFRunLoopCommonModes,
		0,
		&inputData.queue
	);
	
	if (status != noErr) {
		NSLog(@"AudioQueueNewInput returned %d", status);
	}
	
	//Get the full audio format from an audio queue, p. 30
	
	UInt32 size = sizeof inputData.format;
	status = AudioQueueGetProperty(
		inputData.queue,
		kAudioQueueProperty_StreamDescription,
		&inputData.format,
		&size
	);
	
	if (status != noErr) {
		NSLog(@"AudioQueueGetProperty returned %d", status);
	}
	
	//Create an audio file.

	status = AudioFileCreateWithURL(
		url,
		fileType,
		&inputData.format,
		kAudioFileFlags_EraseFile,
		&inputData.fileID
	);
	
	if (status != noErr) {
		NSLog(@"AudioFileCreateWithURL returned %d", status);
	}
	
	//Set an audio queue buffer size, p. 32
	
	inputBufferSize(
		inputData.queue,
		&inputData.format,
		.5,	//seconds
		&inputData.bufferByteSize
	);
	
	NSLog(@"inputData.bufferByteSize == %d", inputData.bufferByteSize);
	
	//Prepare a set of audio queue buffers, p. 33
	
	for (size_t i = 0; i < numberInputBuffers; ++i) {
		status = AudioQueueAllocateBuffer(inputData.queue,
			inputData.bufferByteSize, inputData.buffer + i);
		
		if (status != noErr) {
			NSLog(@"AudioQueueAllocateBuffer returned %d", status);
		}
		
		status = AudioQueueEnqueueBuffer(
			inputData.queue,
			inputData.buffer[i],
			0,
			NULL
		);
		
		if (status != noErr) {
			NSLog(@"initWithFrame: AudioQueueEnqueueBuffer returned %d", status);
		}
	}
	
	inputData.currentPacket = 0;
	inputData.isRunning = true;
	status = AudioQueueStart(inputData.queue, NULL);
	if (status != noErr) {
		NSLog(@"AudioQueueStart returned %d", status);
	}
	light.backgroundColor = [UIColor redColor];
	
	//We can play now that we have recorded something.
	[control setEnabled: YES forSegmentAtIndex: 2];
}

- (void) play {
	[self stop];

	OSStatus status = AudioFileOpenURL(
		url,
		kAudioFileReadPermission,
		fileType,
		&outputData.fileID
	);
	if (status != noErr) {
		NSLog(@"AudioFileOpenURL returned %d", status);
	}
	
	UInt32 size = sizeof outputData.format;
	status = AudioFileGetProperty(
		outputData.fileID,
		kAudioFilePropertyDataFormat,
		&size,
		&outputData.format
	);
	if (status != noErr) {
		NSLog(@"AudioFileGetProperty", status);
	}
	//NSLog(@"size == %u", size);
	
	status = AudioQueueNewOutput(
		&outputData.format,
		outputCallback,
		&outputData,
		CFRunLoopGetCurrent(),
		kCFRunLoopCommonModes,
		0,
		&outputData.queue
	);
	if (status != noErr) {
		NSLog(@"AudioQueueNewOutput returned %d", status);
	}
	
	UInt32 maxPacketSize;
	UInt32 propertySize = sizeof maxPacketSize;
	
	status = AudioFileGetProperty(
		outputData.fileID,
		kAudioFilePropertyPacketSizeUpperBound,
		&propertySize,
		&maxPacketSize
	);
	if (status != noErr) {
		NSLog(@"AudioFileGetProperty returned %d", status);
	}
	//NSLog(@"maxPacketSize == %u", maxPacketSize);
	
	outputBufferSize(
		&outputData.format,
		maxPacketSize,
		.5,
		&outputData.bufferByteSize,
		&outputData.numPacketsToRead
	);
	
	NSLog(@"outputData.bufferByteSize == %u",
		  outputData.bufferByteSize);
	NSLog(@"outputData.numPacketsToRead == %u",
		  outputData.numPacketsToRead);
	
	//constant bitrate
	bool isCBR = outputData.format.mBytesPerPacket > 0
		&& outputData.format.mFramesPerPacket > 0;
	
	if (isCBR) {
		outputData.description = NULL;
	} else {
		outputData.description = (AudioStreamPacketDescription *)
			malloc(outputData.numPacketsToRead
			* sizeof (AudioStreamPacketDescription));
		
		if (outputData.description == NULL) {
			NSLog(@"malloc failed");
		}
	}
	
	//Prime the queue with some data before starting.
	outputData.currentPacket = 0;
	outputData.isRunning = true;
	
	for (size_t i = 0; i < numberOutputBuffers; ++i) {
		status = AudioQueueAllocateBuffer(
			outputData.queue,
			outputData.bufferByteSize,
			outputData.buffer + i
		);
		
		if (status != noErr) {
			NSLog(@"AudioQueueAllocateBuffer returned %d", status);
		}
		
		outputCallback(
			&outputData,
			outputData.queue,
			outputData.buffer[i]
		);
	}
	
	status = AudioQueueSetParameter(
		outputData.queue,
		kAudioQueueParam_Volume,
		1.0
	);
	
	if (status != noErr) {
		NSLog(@"AudioQueueSetParameter returned %d", status);
	}

	status = AudioQueueStart(outputData.queue, NULL);
	if (status != noErr) {
		NSLog(@"AudioQueueStart returned %d", status);
	}
	
	do {
		SInt32 result = CFRunLoopRunInMode(
			kCFRunLoopDefaultMode,
			.25,
			false
		);
		
		if (result != kCFRunLoopRunTimedOut) {
			NSLog(@"CFRunLoopRunInMode returned %d", result);
		}
	} while (outputData.isRunning);
	
	SInt32 result = CFRunLoopRunInMode(
		kCFRunLoopDefaultMode,
		1.0,
		false
	);
	
	if (result != kCFRunLoopRunTimedOut) {
		NSLog(@"CFRunLoopRunInMode returned %d", result);
	}
	
	control.selectedSegmentIndex = -1;	//Unpress the play button.
}

- (void) valueChanged: (id) sender {
	if (sender == control) {
		switch ([control selectedSegmentIndex]) {
				
			case 0:	//Record
				[self record];
				break;
				
			case 1:	//Stop
				[self stop];
				break;
				
			case 2:	//Play
				[self play];
				break;
				
			case -1:	//Play button unpressed when playback is done.
				break;
				
			default:
				NSLog(@"bad selectedSegmentIndex %d",
					  [control selectedSegmentIndex]);
				break;
		}
	}
}

/*
// Only override drawRect: if you perform custom drawing.
// An empty implementation adversely affects performance during animation.
- (void) drawRect: (CGRect) rect {
	// Drawing code
}
*/

- (void) dealloc {
	[light release];
	[control release];
	if (url != NULL) {
		CFRelease(url);
	}
	[super dealloc];
}

@end
